/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000, 2001  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <pan/acache.h>
#include <pan/article.h>
#include <pan/debug.h>
#include <pan/gnksa.h>
#include <pan/group.h>
#include <pan/pan-glib-extensions.h>
#include <pan/pan-i18n.h>
#include <pan/server.h>
#include <pan/group.h>
#include <pan/util-mime.h>

const gchar * default_incoming_name_addr;
const gchar * default_incoming_name_real;

static void fire_articles_changed (Group*, Article**, gint article_qty, ArticleChangeType);

/***
****
****  LIFE CYCLE
****
***/

void
article_destructor (PanObject* o)
{
	Article * a = ARTICLE(o);
        g_return_if_fail (a != NULL);

	/* clean up article parts */
	debug1 (DEBUG_PAN_OBJECT, "article_destructor: %p", a);
	if (a->headers != NULL) {
		g_hash_table_destroy (a->headers);
		a->headers = NULL;
	}
	if (a->threads != NULL) {
		g_slist_free (a->threads);
		a->threads = NULL;
	}
	a->date = 0;
	a->number = 0;
	a->linecount = 0;
	a->message_id = NULL;
	a->references = NULL;
	a->subject = NULL;
	a->author_addr = NULL;
	a->author_real = NULL;
	a->group = NULL;

	/* clean up parent */
	pan_object_destructor (o);
}

void
article_constructor (Article * article, Group * group)
{
	g_return_if_fail (article!=NULL);
	g_return_if_fail (group!=NULL);

	/* init the superclass */
	pan_object_constructor (PAN_OBJECT(article), article_destructor);

	/* init this class */
        debug1 (DEBUG_PAN_OBJECT, "article_constructor: %p", article);
	article->group = group;
	article->self_passes_filter = FALSE;
	article->tree_passes_filter = FALSE;
	article->part = (gint8)0;
	article->parts = (gint8)0;
	article->crosspost_qty = (gint8)0;
	article->linecount = (guint16)0;
	article->state = (guint16)0;
	article->unread_children = (guint16)0;
	article->date = (time_t)0;
	article->number = (gulong)0;
	article->subject = NULL;
	article->author_real = NULL;
	article->author_addr = NULL;
	article->message_id = NULL;
	article->references = NULL;
	article->parent = NULL;
	article->threads = NULL;
	article->headers = NULL;
}

Article*
article_new (Group * group)
{
	Article* article = g_new (Article, 1);
	article_constructor (article, group);
	return article;
}


static void
article_dup_ghfunc (gpointer key, gpointer val, gpointer user_data)
{
	Article * article = ARTICLE(user_data);
	article_set_header (article, key, val, DO_CHUNK);
}
Article*
article_dup (Group * group, const Article * article)
{
	Article * retval;

	g_return_val_if_fail (group!=NULL, NULL);
	g_return_val_if_fail (article!=NULL, NULL);

#define chunk_if_exists(A) \
	is_nonempty_string (A) ? group_chunk_string (group, A) : NULL

	retval                      = article_new (group);
	retval->self_passes_filter  = FALSE;
	retval->tree_passes_filter  = FALSE;
	retval->unread_children     = 0;
	retval->group               = group;
	retval->parts               = article->parts;
	retval->part                = article->part;
	retval->linecount           = article->linecount;
	retval->crosspost_qty       = article->crosspost_qty;
	retval->state               = article->state;
	retval->date                = article->date;
	retval->number              = article->number;
	retval->subject             = chunk_if_exists (article->subject);
	retval->author_real         = chunk_if_exists (article->author_real);
	retval->author_addr         = chunk_if_exists (article->author_addr);
	retval->message_id          = chunk_if_exists (article->message_id);
	retval->references          = chunk_if_exists (article->references);
	retval->parent              = NULL;
	retval->threads             = NULL;
	retval->headers             = NULL;

	if (article->headers != NULL)
		g_hash_table_foreach (article->headers, article_dup_ghfunc, retval);

	return retval;
}

/***
****
***/

static void
articles_set_dirty (Article ** articles, gint qty)
{
	gint i;
	Group * group;

	g_return_if_fail (articles != NULL);
	g_return_if_fail (qty >= 1);

	/* flag the group as dirty */
	group = articles[0]->group;
	for (i=0; i<qty; ++i)
		group_set_articles_dirty (group, TRUE);

	fire_articles_changed (group, articles, qty, ARTICLE_CHANGED_DIRTY);
}



/**
 * Remove an article from its parent/child threadings.  The children and
 * parent are linked together instead.  This is useful for not breaking
 * the threading of articles surrounding an article which is about to be
 * destroyed.
 */
void
article_isolate (Article* article)
{
	GSList *l = NULL;
	Article *parent = article->parent;

	if (parent)
		parent->threads = g_slist_remove (parent->threads, article);

	for (l=article->threads; l!=NULL; l=l->next) {
		Article *child = ARTICLE(l->data);
		child->parent = parent;
		if (parent)
			parent->threads = g_slist_append (parent->threads, child);
	}

	g_slist_free (article->threads);
	article->threads = NULL;
	article->parent = NULL;
}


gboolean
article_is_new (const Article * a)
{
	g_return_val_if_fail (a!=NULL, FALSE);
	g_return_val_if_fail (a->group!=NULL, FALSE);
	return a->number > a->group->article_high_old;
}


/***
****
****  READ / UNREAD
****
***/

gboolean
article_is_read (const Article * a)
{
	g_return_val_if_fail (a!=NULL, FALSE);
	g_return_val_if_fail (a->group!=NULL, FALSE);

	return newsrc_is_article_read (group_get_newsrc(a->group), a->number);
}

static void
article_set_unread_impl (Article    * article,
                         GPtrArray  * changed)
{
	const gboolean was_read = newsrc_mark_article (group_get_newsrc(article->group),
	                                               article->number,
	                                               FALSE);
	if (was_read)
		g_ptr_array_add (changed, article);

	/* FIXME: I don't think this special case is such a great idea.
	   Special case: if it's a multipart, mark 1..n as read too */
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList * l;
		for (l=article->threads; l!=NULL; l=l->next) {
			Article * child = ARTICLE(l->data);
			if (child->part > 0)
				article_set_unread_impl (child, changed);
		}
	}
}

static void
article_set_read_impl (Article * article,
		       GPtrArray * changed)
{
	const gboolean was_read = newsrc_mark_article (group_get_newsrc(article->group),
	                                               article->number,
	                                               TRUE);
	if (!was_read)
		g_ptr_array_add (changed, article);

	/* FIXME: I don't think this special case is such a great idea.
	   Special case: if it's a multipart, mark 1..n as read too */
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList * l;
		for (l=article->threads; l!=NULL; l=l->next) {
			Article * child = ARTICLE(l->data);
			if (child->part > 0)
				article_set_read_impl (child, changed);
		}
	}
}

void
articles_set_read (Article ** articles, int article_qty, gboolean read)
{
	int i;
	GPtrArray * changed;
	debug_enter ("articles_set_read");

	g_return_if_fail (articles != NULL);

	/* build an array of articles whose read state changed */
	changed = g_ptr_array_new ();
	pan_g_ptr_array_reserve (changed, article_qty);

	/* change the articles */
	for (i=0; i<article_qty; ++i)
	{
		Article * article = articles[i];

		/* sanity clause */
		g_return_if_fail (article != NULL);
		g_return_if_fail (article->group != NULL);

		if (articles[0]->group != article->group) {
			pan_warn_if_reached ();
			continue;
		}

		if (read)
			article_set_read_impl (article, changed);
		else 
			article_set_unread_impl (article, changed);
	}

	if (changed->len != 0)
	{
		const gint inc = read ? changed->len : -(int)changed->len;
		Group * group = articles[0]->group;
		group_inc_article_read_qty (group, inc);
		fire_articles_changed (group, articles, article_qty, ARTICLE_CHANGED_READ);
	}

	debug_exit ("articles_set_read");
}

void
article_set_read (Article * article, gboolean read)
{
	g_return_if_fail (article!=NULL);

	articles_set_read (&article, 1, read);
}
        
        


/**
*** PUBLIC MUTATORS
**/

void
articles_add_flag (Article  ** articles,
                   gint        qty,
                   gushort     flag)
{
	gint i;

	g_return_if_fail (articles != NULL);
	g_return_if_fail (qty >= 1);

	for (i=0; i<qty; ++i)
		articles[i]->state |= flag;

	articles_set_dirty (articles, qty);
}


void
articles_remove_flag (Article         ** articles,
                      gint               qty,
                      gushort            flag)
{
	gint i;

	g_return_if_fail (articles != NULL);
	g_return_if_fail (qty >= 1);

	for (i=0; i<qty; ++i)
		articles[i]->state &= ~flag;

	articles_set_dirty (articles, qty);
}



/***
****
****  THREADS
****
***/

gchar*
article_get_thread_message_id (const Article* article)
{
	char * pch;
	const gchar * refs = article_get_header (article, HEADER_REFERENCES);
	const gchar * msg_id = article_get_header (article, HEADER_MESSAGE_ID);

	g_return_val_if_fail (article!=NULL, NULL);
	g_return_val_if_fail (is_nonempty_string(msg_id), NULL);

	if (is_nonempty_string(refs))
	{
		pch = get_next_token_str (refs, ' ', NULL);
	}
	else /* top of the thread */
	{
		pch = g_strdup (msg_id);
	}

	return pch;
}

static void
article_get_entire_thread_impl (Article* top, GPtrArray* setme)
{
	GSList *l;
	g_ptr_array_add (setme, top);
	for (l=top->threads; l!=NULL; l=l->next)
		article_get_entire_thread_impl (ARTICLE(l->data), setme);
}

static void
article_get_thread_impl (Article * article,
                         GPtrArray * setme,
                         gboolean children_only)
{
	Article* top = article;

	if (top != NULL)
	{
		if (!children_only)
			while (top->parent != NULL)
				top = top->parent;

		pan_g_ptr_array_reserve (setme, 128);
		article_get_entire_thread_impl (top, setme);
	}
}

void
article_get_subthread (Article* article, GPtrArray* setme)
{
	article_get_thread_impl (article, setme, TRUE);
}

void
article_get_entire_thread (Article* article, GPtrArray* setme)
{
	article_get_thread_impl (article, setme, FALSE);
}


/***
****
****  OTHER HEADERS
****
***/

void
article_remove_header (Article * a, const gchar * key)
{
	g_return_if_fail (a!=NULL);
	g_return_if_fail (key!=NULL);
		 
	if (!pan_strcmp(key, HEADER_MESSAGE_ID)) {
		a->message_id = NULL;
	} else if (!pan_strcmp(key, HEADER_SUBJECT)) {
		a->subject = NULL;
	} else if (!pan_strcmp(key, HEADER_REFERENCES)) {
		a->references = NULL;
	} else if (!pan_strcmp(key, HEADER_FROM)) {
		a->author_addr = NULL;
		a->author_real = NULL;
	} else if (a->headers!=NULL) {
		g_hash_table_remove (a->headers, key);
		if (!g_hash_table_size(a->headers)) {
			g_hash_table_destroy (a->headers);
			a->headers = NULL;
		}
	}
}

const gchar*
article_get_header (const Article * a, const gchar * key)
{
	const gchar * retval = NULL;

	g_return_val_if_fail (a!=NULL, NULL);
	g_return_val_if_fail (is_nonempty_string(key), NULL);

	if (key == HEADER_MESSAGE_ID) retval = a->message_id;
	else if (key == HEADER_SUBJECT) retval = a->subject;
	else if (key == HEADER_REFERENCES) retval = a->references;
	else if (key == HEADER_FROM) pan_warn_if_reached ();
	else if (!strcmp(key, HEADER_MESSAGE_ID)) retval = a->message_id;
	else if (!strcmp (key, HEADER_SUBJECT)) retval = a->subject;
	else if (!strcmp (key, HEADER_REFERENCES)) retval = a->references;
	else if (!strcmp (key, HEADER_FROM)) pan_warn_if_reached ();
	else if (a->headers!=NULL)
		retval = (const gchar*) g_hash_table_lookup (a->headers, key);

	return retval;
}

gboolean
article_header_is_internal (const gchar * key)
{
	return is_nonempty_string(key) && !g_strncasecmp(key,"Pan-",4);
}

gboolean
article_header_is_extra (const gchar * key)
{
	/* sanity check */
	if (!is_nonempty_string(key)) return FALSE;

	/* pan internals aren't user-specified headers */
	if (article_header_is_internal(key)) return FALSE;

	/* other headers that are handled explicitly elsewhere */
	if (!pan_strcmp(key,HEADER_FOLLOWUP_TO)) return FALSE;
	if (!pan_strcmp(key,HEADER_NEWSGROUPS)) return FALSE;
	if (!pan_strcmp(key,HEADER_ORGANIZATION)) return FALSE;
	if (!pan_strcmp(key,HEADER_REPLY_TO)) return FALSE;

	return TRUE;
}

static void
ensure_extra_headers_exists (Article * a)
{
	if (a->headers == NULL)
		a->headers = g_hash_table_new (g_str_hash, g_str_equal);
}

void
article_set_header (Article         * a,
                    const gchar     * key,
                    const gchar     * val,
                    HeaderAction      action)
{
	g_return_if_fail (a!=NULL);
	g_return_if_fail (is_nonempty_string(key));

	/* chunk if necessary */
	if (action & DO_CHUNK)
	{
		key = group_chunk_string (a->group, key);
		if (val != NULL)
			val = group_chunk_string (a->group, val);
	}

	/* if we're possibly erasing a header, remove the old value */
	if (!is_nonempty_string (val))
		article_remove_header (a, key);

	/* add the new header */
	if (is_nonempty_string (val))
	{
		if      (!pan_strcmp (key, HEADER_MESSAGE_ID)) a->message_id = val;
		else if (!pan_strcmp (key, HEADER_SUBJECT)) a->subject = val;
		else if (!pan_strcmp (key, HEADER_REFERENCES)) a->references = val;
		else if (!pan_strcmp (key, HEADER_FROM)) article_set_author_from_header (a, val);
		else {
			ensure_extra_headers_exists (a);
			g_hash_table_insert (a->headers,
			                     (gpointer)key,
			                     (gpointer)val);
			articles_set_dirty (&a, 1);
		}
	}
}

static void
article_get_all_headers_ghfunc (gpointer key, gpointer value, gpointer user_data)
{
	GPtrArray * a = (GPtrArray*)user_data;
	g_ptr_array_add (a, key);
	g_ptr_array_add (a, value);
}

GPtrArray*
article_get_all_headers (const Article * a)
{
	GPtrArray * retval = g_ptr_array_new ();

	/* sanity clause */
	g_return_val_if_fail (a!=NULL, retval);

	/* add the headers */
	if (a->headers != NULL)
	{
		pan_g_ptr_array_reserve (retval, g_hash_table_size(a->headers)*2);
		g_hash_table_foreach (a->headers, article_get_all_headers_ghfunc, retval);
	}

	return retval;
}

gboolean
article_has_attachment (const Article *a)
{
	return article_get_header(a,PAN_ATTACH_FILE) != NULL;
}

/***
****
****  BODY
****
***/

gboolean
article_has_body (const Article * a)
{
	gboolean retval = FALSE;

	/* maybe the article's in a folder so the body
	 * is hidden as a header... */
	if (!retval) {
		const gchar * body = article_get_header (a, PAN_BODY);
		if (body != NULL)
			retval = TRUE;
	}

	/* try to get it from acache... */
	if (!retval && is_nonempty_string(a->message_id))
		retval = acache_has_message (a->message_id);

	return retval;
}

gchar*
article_get_author_str (const Article * a)
{
        gchar * retval;

	if (a==NULL)
	{
		retval = g_strdup ("");
	}
	else if (is_nonempty_string(a->author_addr))
	{
		if (is_nonempty_string(a->author_real))
		{
			retval = g_strdup_printf ("\"%s\" <%s>", a->author_real, a->author_addr);
		}
		else
		{
			retval = g_strdup (a->author_addr);
		}
	}
	else if (is_nonempty_string(a->author_real))
	{
		retval = g_strdup (a->author_real);
	}
	else
	{
		retval = g_strdup ("");
	}

	g_strstrip (retval);

	return retval;
}

gchar*
article_get_short_author_str (const Article * a)
{
	gchar * author;

	if (is_nonempty_string (a->author_real) && strcmp(a->author_real,default_incoming_name_real))
	{
		author = g_strdup (a->author_real);
	}
	else if (is_nonempty_string (a->author_addr) && strcmp(a->author_addr,default_incoming_name_addr))
	{
		author = g_strdup (a->author_addr);
		if (author != NULL) {
			gchar * pch = strchr (author, '@');
			if (pch != NULL)
				*pch = '\0';
		}
	}
	else
	{
		author = g_strdup (default_incoming_name_real);
	}

	return author;
}


gchar*
article_get_message (const Article * a)
{
	gchar * retval = NULL;

	if (retval==NULL && a!=NULL && acache_has_message (a->message_id))
	{
		retval = acache_get_message (a->message_id);
	}

	if (retval==NULL && a!=NULL)
	{
		GString * s = g_string_new (NULL);
		gchar * pch;

		pch = article_get_headers (a);
		g_string_append (s, pch);
		g_free (pch);

		g_string_append (s, "\n");

		pch = article_get_body (a);
		g_string_append (s, pch);
		g_free (pch);

		retval = s->str;
		g_string_free (s, FALSE);
	}

	return retval;
}

gchar*
article_get_headers (const Article * a)
{
	gchar * retval = NULL;

	/* do we have it in acache? */
	if (retval==NULL && a!=NULL && acache_has_message (a->message_id))
	{
		gchar * s = acache_get_message (a->message_id);
		gchar * pch = strstr (s, "\n\n");
		if (pch != NULL)
			retval = g_strndup (s, (gint)(pch-s));
		g_free (s);
	}

	/* can we build it from a? */
	if (retval==NULL && a!=NULL)
	{
		gchar * tmp;
		const gchar * c_tmp;
		GString * s = g_string_new (NULL);

		/* subject */
		c_tmp = a->subject;
		if (is_nonempty_string (c_tmp))
			g_string_sprintfa (s, _("Subject: %s\n"), c_tmp);

		/* author */
	       	tmp = article_get_author_str (a);
		if (is_nonempty_string (tmp))
			g_string_sprintfa (s, _("From: %s\n"), tmp);
		g_free (tmp);

		/* date */
		if (a->date != 0) {
			tmp = rfc822_date_generate (a->date);
			g_string_sprintfa (s, _("Date: %s\n"), tmp);
			g_free (tmp);
		}

		/* message-id */
		c_tmp = a->message_id;
		if (is_nonempty_string (c_tmp))
			g_string_sprintfa (s, _("Message-Id: %s\n"), c_tmp);

		/* references */
		c_tmp = a->references;
		if (is_nonempty_string (c_tmp))
			g_string_sprintfa (s, _("References: %s\n"), c_tmp);

		/* extra headers */
		if (1) {
			gint i;
			GPtrArray * z = article_get_all_headers (a);
			for (i=0; i<z->len; i+=2) {
				const char* key = (const char*) g_ptr_array_index (z, i);
				const char* val = (const char*) g_ptr_array_index (z, i+1);
				if (!article_header_is_internal (key))
					g_string_sprintfa (s, "%s: %s\n",  key, val);
			}
			g_ptr_array_free (z, TRUE);
		}

		retval = s->str;
		g_string_free (s, FALSE);
	}

	return retval;
}

gchar*
article_get_body (const Article * a)
{
	gchar * retval = NULL;

	/* see if we've got the body hidden in a header... */
	if (retval==NULL && a!=NULL)
	{
		const gchar * body = article_get_header (a, PAN_BODY);
		if (body != NULL)
			retval = g_strdup (body);
	}

	/* see if we've got it in acache... */
	if (retval==NULL && a!=NULL && acache_has_message(a->message_id))
	{
		gchar * s = acache_get_message (a->message_id);
		gboolean is_html = FALSE;
		GMimeMessage * m = pan_g_mime_parser_construct_message (s);
		retval = g_mime_message_get_body (m, TRUE, &is_html);
		g_mime_message_destroy (m);
		g_free (s);
	}

	return retval;
}

void
article_set_author_from_header (Article * a, const gchar * header_from)
{
	gchar * pch;
	gchar * author_addr = NULL;
	gchar * author_real = NULL;

	g_return_if_fail (a!=NULL);
	g_return_if_fail (a->group!=NULL);
	g_return_if_fail (is_nonempty_string(header_from));

	/* note that we don't check gnksa's return value here,
	   because often posted messages on usenet are 
	   syntactically invalid */
	gnksa_do_check_from (header_from, &author_addr, &author_real);

	/* use the real mail address, or fill in a default */
	pch = author_addr;
	if (!is_nonempty_string (pch))
		pch = (gchar*)default_incoming_name_addr;
	a->author_addr = group_chunk_string (a->group, pch);

	/* use the real name, or fill in a default. */
	if (is_nonempty_string(author_real)) {
		gnksa_strip_realname (author_real);
		a->author_real = group_chunk_string (a->group, author_real);
	}
	else if ((pch = strchr(author_addr,'@')) != NULL) {
		gchar * tmp = g_strndup (author_addr, pch-author_addr);
		a->author_real = group_chunk_string (a->group, tmp);
		g_free (tmp);
	} else {
		const gchar * cpch = (gchar*)default_incoming_name_real;
		a->author_real = group_chunk_string (a->group, cpch);
	}

	articles_set_dirty (&a, 1);

	/* cleanup */
	g_free (author_addr);
	g_free (author_real);
}


void
article_set_from_raw_message   (Article         * a,
                                const gchar     * text)
{
	guint i;
	guint16 lines;
	GMimeMessage * msg;
	GPtrArray * arr;
	gchar * body;
	const gchar * pch;
	gboolean foo;

	/* get the message */
	g_return_if_fail (a != NULL);
	g_return_if_fail (is_nonempty_string(text));
	msg = pan_g_mime_parser_construct_message (text);
	g_return_if_fail (msg != NULL);

	/* body */
	a->date = msg->header->date;
	body = g_mime_message_get_body (msg, TRUE, &foo);
	article_set_header (a, PAN_BODY, body, DO_CHUNK);
	for (lines=0, pch=body; pch!=NULL && *pch; ++pch)
		if (*pch=='\n')
			++lines;
	a->linecount = lines;
	g_free (body);

	/* headers */
	article_set_header (a, HEADER_FROM, msg->header->from, DO_CHUNK);
	article_set_header (a, HEADER_REPLY_TO, msg->header->reply_to, DO_CHUNK);
	article_set_header (a, HEADER_SUBJECT, msg->header->subject, DO_CHUNK);
	article_set_header (a, HEADER_MESSAGE_ID, msg->header->message_id, DO_CHUNK);
	arr = msg->header->arbitrary_headers;
	for (i=0; i!=arr->len; ++i) {
		GMimeHeader * h = (GMimeHeader*) g_ptr_array_index (arr, i);
		const gchar * name = h->name;
		if (!pan_strcmp (name, "Author")) /* back compatability with 0.9.2 */
			name = HEADER_FROM;
		article_set_header (a, h->name, h->value, DO_CHUNK);
	}

	/* state */
	pch = article_get_header (a, "Status");
	if (pch != NULL) {
		const gboolean read = strchr (pch, 'R') != NULL;
		article_set_read (a, read);
		/* const gboolean old = strchr (pch, 'O') != NULL; */
		article_remove_header (a, "Status");
	}

	/* references */
	pch = article_get_header (a, HEADER_REFERENCES);
	if (pch != NULL)
		a->references = pch;

	/* cleanup */
	g_mime_message_destroy (msg);
}



/***
****  EVENTS
***/

PanCallback*
article_get_articles_changed_callback (void)
{
	static PanCallback * cb = NULL;
	if (cb==NULL) cb = pan_callback_new ();
	return cb;
}

static void
fire_articles_changed (Group              * group,
                       Article           ** articles,
                       gint                 article_qty,
                       ArticleChangeType    type)
{
	ArticleChangeEvent e;

	g_return_if_fail (group!=NULL);
	g_return_if_fail (articles!=NULL);
	g_return_if_fail (article_qty>0);
	g_return_if_fail (type==ARTICLE_CHANGED_READ || type==ARTICLE_CHANGED_DIRTY);

	e.group = group;
	e.articles = articles;
	e.article_qty = article_qty;
	e.type = type;
	pan_callback_call (article_get_articles_changed_callback(), &e, NULL);

	debug_exit ("fire_articles_changed");
}
