/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Author: Charles Kerr <charles@rebelbase.com>
 *
 * Copyright (C) 1999, 2000, 2001  Pan Development Team <pan@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <glib.h>

#include <gmime/gmime.h>

#include <pan/acache.h>
#include <pan/debug.h>
#include <pan/decode.h>
#include <pan/log.h>
#include <pan/pan-i18n.h>
#include <pan/pan-glib-extensions.h>
#include <pan/util-file.h>
#include <pan/util-mime.h>

/**
*** Private Routines
**/

static gchar*
create_filename (const gchar       * path,
                 const gchar       * subject,
                 const gchar       * default_filename,
                 const GMimePart   * part)
{
	gchar * retval;
	GString * filename;
	debug_enter ("create_filename");

	g_return_val_if_fail (is_nonempty_string(path), NULL);
	g_return_val_if_fail (part!=NULL, NULL);

       	filename = g_string_new (NULL);

	/* first try the filename specified by the user */
	if (!filename->len && is_nonempty_string(default_filename))
		g_string_assign (filename, default_filename);

	/* otherwise try the filename specified by the article */
	if (!filename->len) {
		const gchar * pch = g_mime_part_get_filename (part);
		if (is_nonempty_string(pch))
			g_string_assign (filename, pch);
	}

	/* otherwise try the article's subject */
	if (!filename->len && is_nonempty_string(subject))
		g_string_assign (filename, subject);

	/* otherwise punt */
	if (!filename->len)
		g_string_assign (filename, _("UNKNOWN"));

	/* filter out directory characters */
	if (1) {
		const gchar * pch;
		GString * tmp = g_string_sized_new (filename->len);
		for (pch=filename->str; *pch; ++pch) {
			if (*pch==G_DIR_SEPARATOR)
				g_string_append_c (tmp, '_');
			else if (*pch=='\\')
				g_string_append (tmp, "\\\\");
			else
				g_string_append_c (tmp, *pch);
		}
		g_string_assign (filename, tmp->str);
		g_string_free (tmp, TRUE);
	}

	/* add path */
	if (is_nonempty_string(path)) {
		g_string_prepend_c (filename, G_DIR_SEPARATOR);
		g_string_prepend (filename, path);
	}

	/* make sure we've got a unique file */
	if (file_exists (filename->str))
	{
		gint i;
		const gchar * lastdot;
		gchar * lead;
		gchar * tail;

		/* where to add the number? */
		lastdot = strrchr (filename->str, '.');
		if (lastdot != NULL) {
			lead = g_strndup (filename->str, lastdot-filename->str);
			tail = g_strdup (lastdot);
		} else {
			lead = g_strdup (filename->str);
			tail = g_strdup ("");
		}

		/* loop until we get a filename we can use */
		i = 2;
		for (;;) {
			g_string_sprintf (filename, "%s.%d%s", lead, i++, tail);
			if (!file_exists (filename->str))
				break;
		}

		/* cleanup */
		g_free (lead);
		g_free (tail);
	}

	retval = filename->str;
	g_string_free (filename, FALSE);
	pan_normalize_filename_inplace (retval);
	g_message (_("Using filename [%s]"), retval);
	debug_exit ("create_filename");
	return retval;
}

static void
write_big_article (GPtrArray * articles, FILE * fp_out)
{
	gint i;
	debug_enter ("write_big_article");

	/* sanity clause */
	g_return_if_fail (articles!=NULL);
	g_return_if_fail (articles->len);

	/* get the articles & find out how big the buffer must be */
	for (i=0; i!=articles->len; ++i)
	{
		const Article * a;
		gchar * message = NULL;
		gchar * body;

		/* get the article message */
		a = ARTICLE(g_ptr_array_index(articles,i));
		if (a != NULL)
			message = acache_get_message (a->message_id);
		if (message == NULL) {
			log_add_va (LOG_ERROR, _("Couldn't find body for message `%s' - decoding may be incomplete or incorrect"), a->subject);
			continue;
		}

		/* find the beginning of the body */
		body = strstr (message, "\n\n");
		if (body != NULL)
			body += 2;

		/* keep the headers for the first item */
		if (i==0 && body!=NULL)
			fwrite (message, sizeof(char), body-message, fp_out);

		/* if we couldn't find the headers, assume they're missing.
		   AFAIK this won't happen; this is just CYA. */
		if (body == NULL)
			body = message;

		if (strstr (body, "Cut Here -- cut here") == NULL)
		{
			fwrite (body, sizeof(char), strlen(body), fp_out);
		}
		else
		{
			const gboolean ya_newswatcher_fucking_blows_me = TRUE;
			const gchar * march = body;
			const gchar * test_begin = "BEGIN -- Cut Here -- cut here";
			const gchar * test_end = "END -- Cut Here -- cut here";
			const gint test_begin_len = strlen(test_begin);
			const gint test_end_len = strlen(test_end);

			gint line_len = 0;
			const gchar * line_begin = NULL;
			gboolean ignoring = FALSE;

			while (ya_newswatcher_fucking_blows_me && get_next_token_range(march,'\n',&march,&line_begin,&line_len))
			{
				gboolean skip_current = FALSE;

				if (line_len==test_begin_len && !strncmp(line_begin,test_begin,test_begin_len)) {
					skip_current = TRUE;
					ignoring = FALSE;
				}
				else if (line_len==test_end_len && !strncmp(line_begin,test_end,test_end_len)) {
					skip_current = TRUE;
					ignoring = TRUE;
				}
				else if (ignoring)
					skip_current = TRUE;

				if (!skip_current)
					fwrite (line_begin, sizeof(char), line_len+1, fp_out);
			}
		}

		/* cleanup */
		g_free (message);
	}

	debug_exit ("write_big_article");
}

static void
get_array_of_decodable_parts_func (GMimePart * part, gpointer data)
{
	GPtrArray * a = (GPtrArray*) data;
	const GMimeContentType * type = g_mime_part_get_content_type (part);
	if ((is_nonempty_string(g_mime_part_get_filename (part))) ||
		(!g_mime_content_type_is_type (type, "text", "*") && !g_mime_content_type_is_type (type, "multipart", "*")))
		g_ptr_array_add (a, part);
}

static void
get_decodable_parts (GPtrArray       * articles,
                     GMimeMessage   ** setme_msg,
                     GPtrArray      * fillme_array)
{
	FILE * fp;
	gchar * tmpfilename;
	debug_enter ("get_decodable_parts");

	/* sanity check */
	g_return_if_fail (articles!=NULL);

	/* get the message */
	fp = NULL;
	tmpfilename = pan_make_temp (&fp);
	if (fp != NULL)
	{
		/* write the articles to one big file */
		write_big_article (articles, fp);
		fclose (fp);
		fp = NULL;

		/* parse that file */
		fp = fopen (tmpfilename, "r");
		*setme_msg = pan_g_mime_parser_construct_message_from_file (fp);
		fclose (fp);
		remove (tmpfilename);

		/* build the array of decodable parts */
		g_mime_message_foreach_part (*setme_msg, get_array_of_decodable_parts_func, fillme_array);
	}

	g_free (tmpfilename);
	debug_exit ("get_decodable_parts");
}

/**
*** Public Routines
**/

void
decode_article (const decode_data * dd)
{
	static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
	GPtrArray * articles;
	GMimeMessage * mm;
	GPtrArray * attachments;
	GSList * p;
	gint i = 0;
	gboolean use_mutex;
	gboolean success = TRUE;
	Article * first;
	debug_enter ("decode_article");

	/* sanity clause */
	g_return_if_fail (dd!=NULL);
	g_return_if_fail (dd->server!=NULL);
	g_return_if_fail (dd->articles!=NULL);
	for (p=dd->articles; p!=NULL; p=p->next) {
		Article * a = ARTICLE(p->data);	
		g_return_if_fail (a!=NULL);
		g_return_if_fail (is_nonempty_string(a->message_id));
		g_return_if_fail (acache_has_message(a->message_id));
	}

	first = ARTICLE(dd->articles->data);

	/* This is sad, really. individual GMimeMessages are safe enough;
	   the reason we use mutexes here is that decoding a big mp3 takes
	   up so _much_ memory that running two at once makes my system
	   swap to its knees. */
	use_mutex = g_slist_length (dd->articles) > 3;
	if (use_mutex) {
		status_item_emit_status_va (dd->item, _("Waiting to decode \"%s\""), first->subject);
		g_static_mutex_lock (&mutex);
	}

	/* let the client know what we're doing */
	status_item_emit_status_va (dd->item, _("Processing \"%s\""), first->subject);

	/* make a GPtrArray of the articles */
	articles = g_ptr_array_new ();
	for (p=dd->articles; p!=NULL; p=p->next)
		g_ptr_array_add (articles, p->data);

	/* get the decodable parts */
	status_item_emit_status_va (dd->item, _("Decoding \"%s\""), first->subject);
	mm = NULL;
	attachments = g_ptr_array_new ();
	get_decodable_parts (articles, &mm, attachments);

	/* decode & save the parts */
	success = attachments->len != 0;
	for (i=0; success && i!=attachments->len; ++i)
	{
		guint len;
		const gchar * content;
		FILE * fp = NULL;
		const GMimePart * part = (const GMimePart*) g_ptr_array_index (attachments, i);
		gchar * filename = create_filename (dd->path, first->subject, dd->filename, part);

		/* make sure there's content to write */
		content = g_mime_part_get_content (part, &len);
		if (content==NULL || len==0)
			success = FALSE;

		/* try to open a file for writing */
		if (success)
		{
			gchar * path = g_dirname (filename);
			directory_check (path);
			g_free (path);
			fp = fopen (filename, "w+");
			if (fp == NULL)
				success = FALSE;
		}

		/* write the content */
		if (success) {
			fwrite (content, sizeof(char), len, fp);
			fclose (fp);
			fp = NULL;
			status_item_emit_status_va (dd->item, _("Saved \"%s\""), filename);
		}

		/* log the success */
		if (success) {
			gchar * msg = g_strdup_printf (
				_("Decoded \"%s\" from group \"%s\", \"%s\" part #%d"),
				filename,
				(dd->group ? dd->group->name : "Unknown"),
				first->subject, i+1);
			log_add (LOG_INFO, msg);
			debug0 (DEBUG_DECODE, msg);
			g_free (msg);
		}

		/* remember the filename */
		if (success && !article_get_header (first, PAN_HEADER_FILENAME))
			article_set_header (first, PAN_HEADER_FILENAME, filename, DO_CHUNK);

		/* cleanup this iteration */
		g_free (filename);
	}

	/* update the node */
	articles_remove_flag (&first, 1, STATE_DECODE_QUEUED);
	articles_add_flag (&first, 1, (success?STATE_DECODED:STATE_DECODE_FAILED));

	/* cleanup */
	g_ptr_array_free (articles, TRUE);
	g_mime_message_destroy (mm);
	g_ptr_array_free (attachments, TRUE);
	if (use_mutex)
		g_static_mutex_unlock (&mutex);
	debug_exit ("decode_article");
}
