/*
 * Author: Charles Kerr <charles@rebelbase.com>
 *
 * Copyright (C) 1999, 2000, 2001  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-util.h>

#include <pan/base-prefs.h>
#include <pan/log.h>
#include <pan/pan-i18n.h>
#include <pan/pan-glib-extensions.h>
#include <pan/util-file.h>

/***
****
****  TEMP FILES
****
***/


gchar *
pan_make_temp (FILE ** setme_fp)
{
	char * retval = NULL;
	char * fname = g_strdup_printf ("%s/pan%uXXXXXX", temp_dir, (guint)getpid());
	int fd = mkstemp (fname);

	if (fd == -1)
		g_free (fname);
	else {
		retval = fname;
		*setme_fp = fdopen (fd, "w+");
	}

	return retval;
}

void
pan_clean_temp_dir (void)
{
	GString *filename = g_string_new (NULL);
	char *dirpath = g_strdup_printf ("/%s", temp_dir); 
	char *prefix = g_strdup_printf ("pan%u", (unsigned int)getpid ());
	gint prefix_l = strlen (prefix);
	DIR *dir_p = opendir(dirpath);
	struct dirent *dirent_p;

	while ((dirent_p = readdir(dir_p)))
	{
		if (!strncmp (prefix, dirent_p->d_name, prefix_l))
		{
			g_string_sprintf (filename,
					  "/%s/%s", temp_dir, dirent_p->d_name);
			unlink (filename->str);
		}	
	}
	closedir (dir_p);
	g_free (prefix);
	g_free (dirpath);
	g_string_free (filename, TRUE);
}


/***
****
****  FILES
****
***/


gchar*
pan_normalize_filename_inplace (gchar * filename)
{
	gchar * in;
	gchar * out;

	for (in=out=filename; *in; )
		if (in[0]=='/' && in[1]=='/')
			++in;
		else
			*out++ = *in++;
	*out = '\0';

	return filename;
}


gboolean
file_exists (const char* filename)
{
	return is_nonempty_string(filename) && g_file_exists (filename);
}

GArray*
read_file (const char* filename)
{
	GArray* array = NULL;
	FILE* fp;

	/* get length of file... */
	size_t len = get_filesize(filename);
	if (len<=0)
		return NULL;

	/* open the file... */
	fp = fopen ( filename, "r" );
	if (!fp)
		return NULL;

	/* read the file... */
	array = g_array_new (TRUE, FALSE, 1);
	g_array_set_size (array, len);
	array->len = fread (array->data, 1, len, fp);
	fclose (fp);
	return array;
}

size_t
get_filesize (const char* filename)
{
	struct stat buf;
	if ( !stat(filename, &buf) )
		return buf.st_size;

	return -1;
}

void
directory_check (const char *pathname)
{
	g_return_if_fail (is_nonempty_string(pathname));

	if (!g_file_test (pathname, G_FILE_TEST_ISDIR))
	{
		const gchar * in = pathname;
		gchar * buf = g_malloc (strlen(pathname) + 1);
		gchar * out = buf;

		for (;;)
		{
			*out++ = *in;

			if (*in=='\0' || *in==G_DIR_SEPARATOR)
			{
				if (*in==G_DIR_SEPARATOR)
					*out = '\0';

				if (!g_file_test (buf, G_FILE_TEST_ISDIR))
				{
					if (mkdir (buf, S_IRUSR | S_IWUSR | S_IXUSR))
					{
						log_add_va (LOG_ERROR, _("Couldn't create directory `%s': %s"),
						                       buf, g_strerror (errno));
						break;
					}
				}
			}

			if (!*in)
				break;

			++in;
		}

		g_free (buf);
	}
}

/***
****
****  MBOX
****
***/

const gchar*
mbox_get_next_message (const gchar * buf,
                       gchar ** setme_from,
                       gchar ** setme_msg)
{
	GString * tmp = g_string_new (NULL);
	const gchar * march;
	int lines = 0;

	/* find the end of the message */
	march = buf;
	for (;;)
	{
		gchar * pch;
		gchar * line;
		gboolean is_nested_from = FALSE;
		gboolean is_from;

		/* ran out of text */
		if (march == NULL)
			break;

		/* get next line */
		pch = strchr (march, '\n');
		if (pch) {
			line = g_strndup (march, pch-march);
			march = pch + 1;
		} else { /* this new line is the last one */
			line = g_strdup (march);
			march = NULL;
		}

		/* check the line for delimiter */
		is_from = !strncmp (line, "From ", 5);
		if (is_from && lines!=0) {
			g_free (line);
			break;
		}

		/* is this a nested From_ line? */
		if (*line=='>') {
			gchar * pch = line;
			while (*pch=='>')
				++pch;
			if (!strncmp(pch,"From ", 5))
				is_nested_from = TRUE;
		}

		/* append the line to the output */
		if (is_from)
			*setme_from = g_strdup (line);
		else {
			g_string_append (tmp, is_nested_from?line+1:line);
			g_string_append_c (tmp, '\n');
		}

		/* cleanup this iteration */
		g_free (line);
		++lines;
	}

	/* retval & cleanup */
	*setme_msg = tmp->len ? tmp->str : NULL;
	g_string_free (tmp, FALSE);

	return march!=buf ? march : NULL;
}

gchar*
mbox_format_message (const gchar * message,
                     const gchar * author_address,
                     time_t date)
{
	gchar * retval;
	GString * tmp;
	const gchar * march;

	/* default author address */
	if (!is_nonempty_string(author_address)) {
		pan_warn_if_reached ();
		author_address = _("unknown@spammesenseless.com");
	}

	/* sanity clause */
	g_return_val_if_fail (message!=NULL, NULL);
	g_return_val_if_fail (author_address!=NULL, NULL);

	/* add the From_ line.  note ctime() adds the '\n' */
       	tmp = g_string_new (NULL);
	g_string_sprintf (tmp, "From %s %s", author_address, ctime(&date));

	/* walk through the message, adding each line, prepending
	   From_ lines with '>' as specified by the mbox format */
	march = message;
	for (;;)
	{
		const gchar * line_begin;
		const gchar * line_end;
		gboolean is_from = FALSE;
		gboolean is_nested_from = FALSE;

		/* ran out of text */
		if (march == NULL)
			break;

		/* delimit the line */
		line_begin = march;
		line_end = strchr (march, '\n');
		if (line_end != NULL)
			march = line_end + 1;
		else { /* this new line is the last one */
			line_end = line_begin + strlen(line_begin);
			march = NULL;
		}

		/* check the line for delimiter */
		is_from = !strncmp (line_begin, "From ", 5);

		/* is this a nested From_ line? */
		if (*line_begin) {
			const gchar * pch = line_begin;
			while (*pch=='>')
				++pch;
			if (!strncmp(pch,"From ", 5))
				is_nested_from = TRUE;
		}

		/* append the line to the output */
		g_string_sprintfa (tmp, "%s%*.*s\n",
			(is_from||is_nested_from) ? ">" : "",
			line_end-line_begin,
			line_end-line_begin,
			line_begin);
	}

	/* mbox message ends with a blank line */
	if (tmp->len != 0)
		g_string_append_c (tmp, '\n');

	/* cleanup */
	retval = tmp->len ? tmp->str : NULL;
	g_string_free (tmp, retval==NULL);

	return retval;
}
